#!/bin/bash
###############################################################################
#
#  balance-cpus
#
#  V0.1	20110520/gj
#  
#   license: 
#    - GPLv3 (you know the deal...)
#
#   authors:
#   - Guido J"akel, Guido dot Jaekel at GMX dot DE
#
###############################################################################


get_cpus () { # $1: name of container  or  "" for host 
  local RANGE=$(<$CGROUPPATH/$1/cpuset.cpus)
  RANGE=${RANGE//,/ } # relpace all '/' by ' '
  RANGE=${RANGE//-/ -}	# replaye all '-' by ' -'
  local LASTNUMBER=-1
  for NUMBER in $RANGE; do
    if [[ $NUMBER -lt 0 ]]; then
     NUMBER=$((-NUMBER))
      while [[ $LASTNUMBER -lt $NUMBER ]]; do
        EXPAND+=" $((++LASTNUMBER))"
      done
    else
      EXPAND+=" $NUMBER"
    fi
    LASTNUMBER=$NUMBER
  done
  echo $EXPAND	# get rid of the leading white space
}

put_cpus() { # $1: name of container, $2 cpus
  echo "${2// /,}" >$CGROUPPATH/$1/cpuset.cpus
}


cntlist () { # count number of token in $1
  local LIST=${!1}
  if [ -n "$LIST" ]; then
    set $LIST; echo $#
  else
    echo "0"
  fi
}


# get the path to the cgroup-fs (with the cpuset subsys)
CGROUPPATH=$(awk '$3 == "cgroup" &&  $4 ~ /cpuset/  {print $2}' </proc/mounts)
[ ! -e "$CGROUPPATH/cpuset.cpus" ] && echo "FATAL: cgroup-fs not found at $CGROUPPATH!" >&2 && exit 1

# get all container names
CONTAINERS=$(find $CGROUPPATH -type d -printf "%P\n")
NCONTAINERS=$(cntlist CONTAINERS)


# re-balance cpu assignment of all containers
declare -Ai CUR_CONTAINER_CNTS NEW_CONTAINER_CNTS
declare -i NCPUS=0 HOST_NCPUS=0 TOTAL_NCPUS=0 NEW_CPU=0
for CONTAINER in "" $CONTAINERS; do
  CUR_CPUS="$(get_cpus $CONTAINER)"
  NCPUS=$(cntlist CUR_CPUS)
  CHANGE=
  if [ -n "$CONTAINER" ]; then	# a container
    TOTAL_NCPUS+=$NCPUS
    NEW_CPUS=""
    for CUR_CPU in $CUR_CPUS; do
      [[ $((CUR_CONTAINER_CNTS[$CUR_CPU]++)) ]]	# void expression
      # balance cpus
      NEW_CPUS+=" $NEW_CPU"
      [[ $((NEW_CONTAINER_CNTS[$NEW_CPU]++)) ]]
      NEW_CPU=$((++NEW_CPU%HOST_NCPUS))
    done
    NEW_CPUS=${NEW_CPUS:1}
    if [[ "$NEW_CPUS" != "$CUR_CPUS" ]]; then
      put_cpus $CONTAINER "$NEW_CPUS"
      CHANGE="-> ${NEW_CPUS// /,}"
    fi
  else	# the host
    HOST_NCPUS=$NCPUS
  fi
  printf "%-15s %3d  %s %s\n" "${CONTAINER:+ }${CONTAINER:-$HOSTNAME}" $NCPUS "${CUR_CPUS// /,}" "$CHANGE"
done
CPC=$( echo "scale=2; $NCONTAINERS/$HOST_NCPUS" | bc -q 2>/dev/null)

# print cpus info
printf "%d containers claiming %d cpus, commit ratio is %.2f.\n" $NCONTAINERS $TOTAL_NCPUS $CPC
echo -n "prev cpu allocation "; for (( i=0; i<HOST_NCPUS; i++ )); do echo -n "${CUR_CONTAINER_CNTS[$i]:-0} "; done; echo
echo -n "new  cpu allocation "; for (( i=0; i<HOST_NCPUS; i++ )); do echo -n "${NEW_CONTAINER_CNTS[$i]:-0} "; done; echo

